(**

This file formalizes standard Separation Logic, as described in 
Arthur Charguéraud's lecture notes.

This file contains:
- a definition of heaps as finite maps from location to values,
- an instantiation of the functor from the file ModelSepFunctor.v,
- a definition of triples,
- statement and proofs of SL reasoning rules.

Author: Arthur Charguéraud.
License: MIT.

*)

Set Implicit Arguments.
Require Export ModelLambda ModelSepFunctor.
Open Scope state_scope.

Ltac auto_star ::= jauto.


(********************************************************************)
(* * Construction of core of the logic *)

Module SepBasicCore.


(*------------------------------------------------------------------*)
(* ** Types *)

Definition heap : Type := (state)%type.

Definition hprop := heap -> Prop.


(*------------------------------------------------------------------*)
(* ** Properties of heaps *)

(** Definitions used for uniformity with other
    instantiation of the functor *)

Notation "'heap_empty'" := (state_empty : heap) : heap_scope.
Open Scope heap_scope.

Notation "h1 \u h2" := (state_union h1 h2)
   (at level 51, right associativity) : heap_scope.

Definition heap_union_empty_l := state_union_empty_l.
Definition heap_union_empty_r := state_union_empty_r.
Definition heap_union_comm := state_union_comm_disjoint.


(*------------------------------------------------------------------*)
(* ** Operators *)

(* \[] *)

Definition hprop_empty : hprop := 
  fun h => h = heap_empty.

(* H1 \* H2 *)

Definition hprop_star (H1 H2 : hprop) : hprop := 
  fun h => exists h1 h2, H1 h1 
                      /\ H2 h2 
                      /\ (\# h1 h2)
                      /\ h = h1 \+ h2.

(* Hexists x, H *)

Definition hprop_exists A (J : A -> hprop) : hprop := 
  fun h => exists x, J x h.

(* [P] *)

Definition hprop_pure (P:Prop) : hprop := 
  hprop_exists (fun (p:P) => hprop_empty).

(* \GC = Hexists H, H *)

Definition hprop_gc := 
  hprop_exists (fun (H:hprop) => H).


(*------------------------------------------------------------------*)
(* ** Notation *)

Notation "\[]" := (hprop_empty) 
  (at level 0) : heap_scope.
Notation "\[ P ]" := (hprop_pure P) 
  (at level 0, P at level 99) : heap_scope.
Notation "H1 '\*' H2" := (hprop_star H1 H2)
  (at level 41, right associativity) : heap_scope.
Notation "\GC" := (hprop_gc) : heap_scope. 
Open Scope heap_scope.


(*------------------------------------------------------------------*)
(* ** Tactic for automation *)

(* TODO: check how much is really useful *)

Hint Extern 1 (_ = _ :> heap) => state_eq.

Tactic Notation "state_disjoint_pre" :=
  subst; rew_disjoint; jauto_set.

Hint Extern 1 (\# _ _) => state_disjoint_pre.
Hint Extern 1 (\# _ _ _) => state_disjoint_pre.


(*------------------------------------------------------------------*)
(* ** Properties of empty *)

Lemma hprop_empty_prove : 
  \[] heap_empty.
Proof using. hnfs~. Qed.

Lemma hprop_empty_inv : forall h,
  \[] h ->
  h = heap_empty.
Proof using. auto. Qed.


(*------------------------------------------------------------------*)
(* ** Properties of star *)

Section Properties.

Hint Resolve hprop_empty_prove.

Lemma hprop_star_empty_l : forall H,
  hprop_empty \* H = H.
Proof using.
  intros. applys hprop_extens. intros h.
  iff (h1&h2&M1&M2&D&U) M.
  { forwards E: hprop_empty_inv M1. subst.
    rewrite~ heap_union_empty_l. }
  { exists~ heap_empty h. }
Qed.

Lemma hprop_star_comm : forall H1 H2,
   H1 \* H2 = H2 \* H1.
Proof using. 
  intros H1 H2. unfold hprop, hprop_star. extens. intros h.
  iff (h1&h2&M1&M2&D&U); rewrite~ heap_union_comm in U; exists* h2 h1.
Qed.

Lemma hprop_star_assoc : forall H1 H2 H3,
  (H1 \* H2) \* H3 = H1 \* (H2 \* H3).
Proof using. 
  intros H1 H2 H3. applys hprop_extens. intros h. split.
  { intros (h'&h3&(h1&h2&M3&M4&D'&U')&M2&D&U). subst h'.
    exists h1 (h2 \+ h3). splits~. { exists* h2 h3. } }
  { intros (h1&h'&M1&(h2&h3&M3&M4&D'&U')&D&U). subst h'.
    exists (h1 \+ h2) h3. splits~. { exists* h1 h2. } }
Qed. 


(*------------------------------------------------------------------*)
(* ** Interaction of star with other operators *)

Lemma hprop_star_exists : forall A (J:A->hprop) H,
  (hprop_exists J) \* H = hprop_exists (fun x => (J x) \* H).
Proof using.
  intros. applys hprop_extens. intros h. iff M.
  { destruct M as (h1&h2&(x&M1)&M2&D&U). exists~ x h1 h2. }
  { destruct M as (x&(h1&h2&M1&M2&D&U)). exists h1 h2. splits~. exists~ x. }
Qed.

Lemma himpl_frame_l : forall H2 H1 H1',
  H1 ==> H1' -> 
  (H1 \* H2) ==> (H1' \* H2).
Proof using. introv W (h1&h2&?). exists* h1 h2. Qed.


End Properties.

End SepBasicCore.


(********************************************************************)
(* * Properties of the logic *)

Module Export SepBasicSetup := SepLogicSetup SepBasicCore.


(*------------------------------------------------------------------*)
(* ** Singleton heap *)

(** r ~~> v *)

Definition hprop_single (l:loc) (v:val) : hprop := 
  fun h => h = state_single l v /\ l <> null.

Notation "l '~~>' v" := (hprop_single l v)
  (at level 32, no associativity) : heap_scope.

Lemma hprop_star_single_same_loc_disjoint : forall (l:loc) (v1 v2:val),
  (l ~~> v1) \* (l ~~> v2) ==> \[False].
Proof using.
  intros. unfold hprop_single. intros h (h1&h2&E1&E2&D&E). false.
  subst. applys* state_single_same_loc_disjoint.
Qed.

Global Opaque hprop_single.

(* ** Configure [hsimpl] to make it aware of [hprop_single] *)

Ltac hcancel_hook H ::= 
  match H with
  | hprop_single _ _ => hcancel_try_same tt
  end.


(********************************************************************)
(* * Reasoning Rules *)


(*------------------------------------------------------------------*)
(* ** Definition of triples *)

Definition triple t H Q :=
  forall H' h,  
  (H \* H') h -> 
  exists h' v, 
       red h t h' v
    /\ (Q v \* \GC \* H') h'.


(*------------------------------------------------------------------*)
(* ** Structural rules *)

Lemma rule_extract_exists : forall t (A:Type) (J:A->hprop) Q,
  (forall x, triple t (J x) Q) ->
  triple t (hprop_exists J) Q.
Proof using.
  introv M. intros HF h N. rewrite hprop_star_exists in N.
  destruct N as (x&N). applys* M.
Qed.

Lemma rule_extract_prop : forall t (P:Prop) H Q,
  (P -> triple t H Q) ->
  triple t (\[P] \* H) Q.
Proof using.
  intros t. applys (rule_extract_prop_from_extract_exists (triple t)).
  applys rule_extract_exists.
Qed.

Lemma rule_consequence : forall t H' Q' H Q,
  H ==> H' ->
  triple t H' Q' ->
  Q' ===> Q ->
  triple t H Q.
Proof using. 
  introv MH M MQ. intros HF h N.
  forwards (h'&v&R&K): (rm M) HF h. { hhsimpl~. }
  exists h' v. splits~. { hhsimpl~. }
Qed.

Lemma rule_frame : forall t H Q H',
  triple t H Q ->
  triple t (H \* H') (Q \*+ H').
Proof using. 
  introv M. intros HF h N. rewrite hprop_star_assoc in N.
  forwards (h'&v&R&K): (rm M) (H' \* HF) h. { hhsimpl~. }
  exists h' v. splits~. { hhsimpl~. }
Qed.

Lemma rule_gc_post : forall t H Q,
  triple t H (Q \*+ \GC) ->
  triple t H Q.
Proof using.
  introv M. intros HF h N. forwards* (h'&v&R&K): (rm M) HF h.  
  exists h' v. splits~. { rewrite <- hprop_gc_merge. hhsimpl. } 
Qed.

Lemma rule_gc_pre : forall t H Q,
  triple t H Q ->
  triple t (H \* \GC) Q.
Proof using.
  introv M. applys rule_gc_post. applys~ rule_frame.
Qed.


(*------------------------------------------------------------------*)
(* ** Term rules *)

Lemma rule_val : forall v H Q,
  H ==> Q v ->
  triple (trm_val v) H Q.
Proof using.
  introv M. intros HF h N. exists h v. splits~.
  { applys red_val. }
  { hhsimpl. hchanges M. }
Qed.

Lemma rule_if : forall v t1 t2 H Q,
  triple (If v = val_int 0 then t2 else t1) H Q ->
  triple (trm_if v t1 t2) H Q.
Proof using.
  introv M. intros HF h N. forwards* (h'&v'&R&K): (rm M) HF h.
  exists h' v'. splits~. { applys~ red_if. }
Qed.

Lemma rule_let : forall x t1 t2 H Q Q1,
  triple t1 H Q1 ->
  (forall (X:val), triple (subst_trm x X t2) (Q1 X) Q) ->
  triple (trm_let x t1 t2) H Q.
Proof using.
  introv M1 M2. intros HF h N.
  lets~ (h1'&v1&R1&K1): (rm M1) HF h.
  forwards* (h2'&v2&R2&K2): (rm M2) (\GC \* HF) h1'.
  exists h2' v2. splits~.
  { applys~ red_let R2. }
  { rewrite <- hprop_gc_merge. hhsimpl. } 
Qed.

Lemma rule_let_val : forall x v1 t2 H Q,
  (forall (X:val), X = v1 -> triple (subst_trm x X t2) H Q) ->
  triple (trm_let x (trm_val v1) t2) H Q.
Proof using. 
  introv M. forwards~ M': (rm M).
  applys_eq~ (>> rule_let H (fun x => \[x = v1] \* H)) 2.
  { applys rule_val. hsimpl~. }
  { intros X. applys rule_extract_prop. intro_subst. applys M'. }
Qed.

Lemma rule_app : forall f x F V t1 H Q,
  F = (val_fix f x t1) ->
  triple (subst_trm f F (subst_trm x V t1)) H Q ->
  triple (trm_app F V) H Q.
Proof using.
  introv EF M. subst F. intros HF h N.
  lets~ (h'&v&R&K): (rm M) HF h.
  exists h' v. splits~. { applys~ red_app. }
Qed.

Lemma rule_let_fix : forall f x t1 t2 H Q,
  (forall (F:val), 
    (forall X H' Q', 
      triple (subst_trm f F (subst_trm x X t1)) H' Q' ->
      triple (trm_app F X) H' Q') ->
    triple (subst_trm f F t2) H Q) ->
  triple (trm_let f (trm_val (val_fix f x t1)) t2) H Q.
Proof using.
  introv M. applys rule_let_val. intros F EF. 
  applys (rm M). clears H Q. intros X H Q. applys~ rule_app.
Qed.

Section RulesPrimitiveOps.
Transparent hprop_star hprop_single.

Lemma rule_ref : forall v,
  triple (prim_ref v) \[] (fun r => Hexists l, \[r = val_loc l] \* l ~~> v).
Proof using.
  intros. intros HF h N.
  forwards~ (l&Dl&Nl): (state_disjoint_new null h v).
  sets h1': (state_single l v).
  exists (h1' \u h) (val_loc l). splits~.
  { applys~ red_ref. }
  { exists h1' h. split. 
    { exists l. unfold h1'. applys~ himpl_inst_prop (l ~~> v). hnfs~. }
    { splits~. hhsimpl~. } }
Qed.

Lemma rule_get : forall v l,
  triple (prim_get (val_loc l)) (l ~~> v) (fun x => \[x = v] \* (l ~~> v)).
Proof using.
  intros. intros HF h N. exists h v. splits~.
  { applys red_get. destruct N as (?&?&(?&?)&?&?&?).
    subst h. applys~ state_union_single_read. }
  { rew_heap. rewrite hprop_star_pure. split~. hhsimpl~. }
Qed.

Lemma rule_set : forall w l v,
  triple (prim_set (val_pair (val_loc l) w)) (l ~~> v) (fun r => \[r = val_unit] \* l ~~> w).
Proof using.
  intros. intros HF h N. destruct N as (h1&h2&(N0&N1)&N2&N3&N4).
  hnf in N1. sets h1': (state_single l w).
  exists (h1' \u h2) val_unit. splits~.
  { applys red_set. subst h h1. applys~ state_union_single_write. }
  { rew_heap. rewrite hprop_star_pure. split~.
    { exists h1' h2. splits~.
      { hnfs~. }
      { hhsimpl~. }
      { subst h1. applys~ state_disjoint_single_set v. } } }
Qed.

End RulesPrimitiveOps.


(********************************************************************)
(* * Bonus *)

(*------------------------------------------------------------------*)
(* ** Triples satisfy the [local] predicate *)

Lemma is_local_triple : forall t,
  is_local (triple t).
Proof using. 
  intros. applys prop_ext_2. intros H Q. iff M.
  { intros h Hh. forwards (h'&v&N1&N2): M \[] h. { hhsimpl. } 
    exists H \[] Q. splits~. { hhsimpl. } { hsimpl. } }
  { intros H' h Hh. lets (h1&h2&N1&N2&N3&N4): Hh. hnf in M.
    lets (H1&H2&Q1&R1&R2&R3): M N1.
    forwards (h'&v&S1&S2): R2 (H2\*H') h.
    { subst h. rewrite <- hprop_star_assoc. exists~ h1 h2. }
    exists h' v. splits~. rewrite <- hprop_gc_merge.
    applys himpl_forward S2.
    hchange (R3 v). rew_heap.
    rewrite (hprop_star_comm_assoc \GC H'). hsimpl. }  
Qed.


(*------------------------------------------------------------------*)
(* ** Alternative, low-level definition of triples *)

Definition triple' t H Q :=
  forall h1 h2, 
  \# h1 h2 -> 
  H h1 -> 
  exists h1' h3' v, 
       \# h1' h2 h3'
    /\ red (h1 \u h2) t (h1' \u h3' \u h2) v
    /\ (Q v) h1'.

Lemma triple_eq_triple' : triple = triple'.
Proof using.
  hint hprop_gc_prove.
  applys prop_ext_3. intros t H Q.
  unfold triple, triple'. iff M.
  { introv D P1.
    forwards~ (h'&v&R1&R2): M (=h2) (h1 \u h2).
    { exists~ h1 h2. } 
    rewrite <- hprop_star_assoc in R2.
    destruct R2 as (h1''&h2'&N0&N1&N2&N3). subst h2'.
    destruct N0 as (h1'&h3'&T0&T1&T2&T3). 
    exists h1' h3' v. splits~. { state_red. } }
  { introv (h1&h2&N1&N2&D&U).
    forwards~ (h1'&h3'&v&R1&R2&R3): M h1 h2.
    exists (h1' \u h3' \u h2) v. splits~.
    { state_red. } 
    { exists~ h1' (h3' \u h2). splits~.
      exists h3' h2. splits~. } }
Qed.

